# -*- coding: utf-8 -*-
from datetime import datetime, time
from django.utils.timezone import get_default_timezone as gdtz

from django.contrib.auth.models import Permission
from django.db.models.query import Q

from juntagrico.models import Member
from juntagrico.util.temporal import start_of_business_year

from juntagrico_assignment_request.entity.assignment_request import AssignmentRequest


class AssignmentRequestDao:

    @staticmethod
    def all_approvers():
        perm = Permission.objects.get(codename='can_confirm_assignments')
        return Member.objects.filter(Q(user__groups__permissions=perm) |
                                     Q(user__user_permissions=perm)).distinct()

    @staticmethod
    def all_notified_on_unapproved_assignments():
        perm = Permission.objects.get(codename='notified_on_unapproved_assignments')
        return Member.objects.filter(Q(user__groups__permissions=perm) |
                                     Q(user__user_permissions=perm)).distinct()

    @staticmethod
    def current_requests_by_member(member):
        start = gdtz().localize(datetime.combine(start_of_business_year(), time.min))
        return AssignmentRequest.objects.filter(member=member).\
            filter(Q(status=AssignmentRequest.REQUESTED) | Q(job_time__gte=start))



from datetime import datetime, date
import time

from django.db import models
from django.utils.translation import gettext as _
from django.core.validators import MinValueValidator

from juntagrico.entity.member import Member
from juntagrico.entity.jobs import Assignment, ActivityArea, JobType, RecuringJob

from juntagrico.config import Config


class AssignmentRequest(models.Model):
    """
    A request to get an assignment
    """

    JOB_NAME_PREFIX = '_assignment_request_'

    REQUESTED = 'RE'
    REJECTED = 'NO'
    CONFIRMED = 'CO'
    REQUEST_STATUS = [
        (REQUESTED, _('Beantragt')),
        (REJECTED, _('Abgelehnt')),
        (CONFIRMED, _('Bestätigt')),
    ]

    member = models.ForeignKey(Member, verbose_name=Config.vocabulary('member'), on_delete=models.CASCADE,
                               help_text=_("Beantragt von"))
    assignment = models.OneToOneField(Assignment, verbose_name=Config.vocabulary('assignment'),
                                      blank=True, null=True, on_delete=models.PROTECT)
    amount = models.PositiveIntegerField(_('Wert'), default=1, validators=[MinValueValidator(1)],
                                         help_text=_("Wieviele Böhnli?"))
    job_time = models.DateTimeField(_('Geleistet am'), default=datetime.now)
    request_date = models.DateField(_('Beantragt am'), default=date.today, blank=True, null=True)
    response_date = models.DateField(_('Beantwortet am'), blank=True, null=True,
                                     help_text=_("Typischerweise heute"))
    approver = models.ForeignKey(Member, verbose_name=_('Abgesprochen mit'), related_name=_('Referenz'),
                                 blank=True, null=True, on_delete=models.SET_NULL,
                                 help_text=_("An wen soll die Anfrage gesendet werden oder wer hat "
                                             "dich aufgefordert, die Anfrage zu senden?"))

    description = models.TextField(_('Beschreibung'), max_length=1000, default='',
                                   help_text=_("Kurze beschreibung was du gemacht hat"))
    activityarea = models.ForeignKey(ActivityArea, verbose_name=_('Tätigkeitsbereich'),
                                     blank=True, null=True, on_delete=models.SET_NULL,
                                     help_text=_("Was am besten passt. Ansonsten leer lassen"))
    duration = models.PositiveIntegerField(_('Dauer in Stunden'), default=4)
    location = models.CharField(_('Ort'), max_length=100, blank=True, default='',
                                help_text=_("Optional"))

    status = models.CharField(max_length=2, choices=REQUEST_STATUS, default=REQUESTED,
                              help_text=_('Hier "Bestätigt" auswählen, sonst zählt das Böhnli nicht!'))
    response = models.TextField(_('Antwort'), blank=True, null=True,
                                help_text=_("Rückmeldung an Böhnli-beantrager. Kann leer bleiben."))

    def __str__(self):
        return _('%s Anfrage #%s') % (Config.vocabulary('assignment'), self.id)

    def is_confirmed(self):
        return self.status == self.CONFIRMED

    def is_rejected(self):
        return self.status == self.REJECTED

    def save(self, **kwargs):
        old_assignment = self.assignment
        super().save(**kwargs)
        # delete assignment object, if request is unconfirmed again
        # this must be done after fully saving the request,
        # because otherwise the assignment is protected by the request
        if old_assignment and not self.assignment:
            self._remove_assignment(old_assignment)

    @classmethod
    def pre_save(cls, sender, instance, **kwds):
        """
        Callback before saving assignment request
        """
        if instance.status == cls.CONFIRMED:
            cls._create_or_update_assignment(instance)
        else:
            instance.assignment = None  # assignment object will be deleted after saving

    def set_activityarea_if_none(self):
        # create/use default activity_area if none specified
        if not self.activityarea:
            self.activityarea, created = ActivityArea.objects.get_or_create(
                name=_('Selbständige Einsätze'),
                defaults={
                    'coordinator': self.approver,
                    'hidden': True
                }
            )

    def get_matching_job_type(self):
        # look for autogenerated job type that matches with request
        values = {x: getattr(self, x) for x in ['activityarea']}
        job_type = JobType.objects.filter(
            name__startswith=self.JOB_NAME_PREFIX,
            **values
        )[:1]
        if job_type:
            return job_type[0]
        # create job type if not exists
        return JobType.objects.create(
            name=f'{self.JOB_NAME_PREFIX}{self.activityarea.id}_{int(time.time())}',  # generate unique job name.
            **values,
            visible=False,
            location='-',  # does not really matter
            default_duration=1,  # overridden in job anyways
            displayed_name=_('Selbständiger Einsatz'),
            description=_('Dies ist ein automatisch erzeugter Einsatz.'),
        )

    def get_matching_job(self):
        # create job with job type if not exists
        return RecuringJob.objects.get_or_create(type=self.get_matching_job_type(),
                                                 time=self.job_time,
                                                 duration_override=self.duration,
                                                 defaults={'slots': 1})[0]

    @classmethod
    def _create_or_update_assignment(cls, instance):
        instance.set_activityarea_if_none()
        # assignment
        if instance.assignment:
            # update if exists:
            # saving the assignment will automatically restore its content. If old job is empty after that, remove it.
            cls._remove_job(instance.assignment.job, instance.assignment.save)

        else:
            # create new if does not exist
            matching_job = instance.get_matching_job()
            if matching_job.free_slots <= 0:
                matching_job.slots += 1 - matching_job.free_slots
                matching_job.save()
            instance.assignment = Assignment.objects.create(member=instance.member,
                                                            job=matching_job, amount=instance.amount)

    @classmethod
    def _remove_assignment(cls, assignment):
        # Delete assignment and job
        cls._remove_job(assignment.job, assignment.delete)

    @classmethod
    def _remove_job(cls, job, before_delete):
        # Delete job, if it has no other assignments
        # the job type is not deleted
        before_delete()
        if job.occupied_places() == 0:
            job.delete()
        elif job.free_slots > 0:  # if job stays, remove emptied slot
            job.slots -= 1

    class Meta:
        verbose_name = _('%s Anfrage') % Config.vocabulary('assignment')
        verbose_name_plural = _('%s Anfragen') % Config.vocabulary('assignment')
        permissions = (
            ('can_confirm_assignments', _('Kann selbständige Arbeitseinsätze bestätigen')),
            ('notified_on_unapproved_assignments', _('Wird über nicht abgesprochene Arbeitseinsätze informiert')),
        )



from django.contrib.auth.models import Permission
from django.test import override_settings, TestCase
from django.utils import timezone
from juntagrico.entity.jobs import ActivityArea
from juntagrico.entity.member import Member

from juntagrico_assignment_request.entity.assignment_request import AssignmentRequest


@override_settings(EMAIL_BACKEND='django.core.mail.backends.locmem.EmailBackend')
class AssignmentRequestTestCase(TestCase):
    def setUp(self):
        self.set_up_member()
        self.set_up_area()

    def assertGet(self, url, code=200, member=None):
        login_member = member or self.member
        self.client.force_login(login_member.user)
        response = self.client.get(url)
        self.assertEqual(response.status_code, code)

    def assertPost(self, url, data=None, code=200, member=None):
        login_member = member or self.member
        self.client.force_login(login_member.user)
        response = self.client.post(url, data)
        self.assertEqual(response.status_code, code)

    @staticmethod
    def create_member(email):
        member_data = {'first_name': 'first_name',
                       'last_name': 'last_name',
                       'email': email,
                       'addr_street': 'addr_street',
                       'addr_zipcode': 'addr_zipcode',
                       'addr_location': 'addr_location',
                       'phone': 'phone',
                       'confirmed': True,
                       }
        member = Member.objects.create(**member_data)
        member.user.set_password('12345')
        member.user.save()
        return member

    def set_up_member(self):
        """
        member
        """
        self.member = self.create_member('member@email.org')
        self.approver = self.create_member('approver@email.org')
        self.approver2 = self.create_member('approver2@email.org')
        self.approver.user.user_permissions.add(
            Permission.objects.get(codename='can_confirm_assignments'))
        self.approver.user.save()
        self.approver2.user.user_permissions.add(
            Permission.objects.get(codename='can_confirm_assignments'))
        self.approver2.user.user_permissions.add(
            Permission.objects.get(codename='notified_on_unapproved_assignments'))
        self.approver2.user.save()

    @staticmethod
    def admin():
        """
        admin member
        """
        admin_data = {'first_name': 'admin',
                      'last_name': 'last_name',
                      'email': 'admin@email.org',
                      'addr_street': 'addr_street',
                      'addr_zipcode': 'addr_zipcode',
                      'addr_location': 'addr_location',
                      'phone': 'phone',
                      'confirmed': True,
                      }
        admin = Member.objects.create(**admin_data)
        admin.user.set_password("123456")
        admin.user.is_staff = True
        admin.user.is_superuser = True
        admin.user.save()
        return admin

    def set_up_area(self):
        """
        area
        """
        area_data = {'name': 'name',
                     'coordinator': self.approver}
        area_data2 = {'name': 'name2',
                      'coordinator': self.approver2,
                      'hidden': True}
        self.area = ActivityArea.objects.create(**area_data)
        self.area2 = ActivityArea.objects.create(**area_data2)

    def assignment_request_data(self, approver=None, for_form=False, approved=False):
        if for_form:
            if approver is None:
                approver = ''
            else:
                approver = approver.pk
        date = timezone.now()
        data = {
            'job_time': date.strftime('%Y-%m-%d %H:%M') if for_form else date,
            'duration': 4,
            'amount': 1,
            'approver': approver,
            'activityarea': self.area.pk if for_form else self.area,
            'location': 'location',
            'description': 'description'
        }
        if not for_form:
            data['member'] = self.member
            if approved:
                data['status'] = AssignmentRequest.CONFIRMED
        return data

    def assignment_response_data(self, decision='submit'):
        """
        :param decision: ['confirm', 'reject', 'submit'(default)]
        """
        return {
            'response': 'response',
            decision: True,
            'amount': 2,
            'activityarea': self.area.pk,
            'location': 'location'
        }



# encoding: utf-8

from django.conf import settings


class AssignmentRequestConfig:
    @staticmethod
    def emails(key):
        if hasattr(settings, 'EMAILS') and key in settings.EMAILS:
            return settings.EMAILS[key]
        return {
            'new_assignment_request_mail': 'assignment_request/mails/new_assignment_request_mail.txt',
            'edited_assignment_request_mail': 'assignment_request/mails/edited_assignment_request_mail.txt',
            'responded_assignment_request_mail': 'assignment_request/mails/responded_assignment_request_mail.txt',
            'confirmed_assignment_request_mail': 'assignment_request/mails/confirmed_assignment_request_mail.txt',
            'rejected_assignment_request_mail': 'assignment_request/mails/rejected_assignment_request_mail.txt',
            'notify_original_approver_mail': 'assignment_request/mails/notify_original_approver_mail.txt',
        }[key]



from django.contrib import admin

from juntagrico_assignment_request.entity.assignment_request import AssignmentRequest
from juntagrico_assignment_request.admins.assignment_request_admin import *


admin.site.register(AssignmentRequest, AssignmentRequestAdmin)



from django.urls import reverse

from juntagrico_assignment_request.entity.assignment_request import AssignmentRequest
from test import AssignmentRequestTestCase


class AdminTests(AssignmentRequestTestCase):

    def test_assignment_request_admin(self):
        ar = AssignmentRequest.objects.create(**self.assignment_request_data(self.approver))
        self.assertGet(reverse('admin:juntagrico_assignment_request_assignmentrequest_change', args=(ar.pk,)),
                       member=self.admin())



from juntagrico_assignment_request.entity.assignment_request import AssignmentRequest
from test import AssignmentRequestTestCase


class ManipulationTests(AssignmentRequestTestCase):
    def setUp(self):
        super().setUp()
        self.ar = AssignmentRequest.objects.create(**self.assignment_request_data(self.approver, approved=True))

    def test_job_type_manipulation(self):
        self.ar.assignment.job.type.activityarea = self.area2
        self.ar.save()
        self.assertEqual(self.ar.assignment.job.type.activityarea, self.area)

    def test_job_manipulation(self):
        self.ar.assignment.job.multiplier = 2
        self.ar.save()
        self.assertEqual(self.ar.assignment.job.multiplier, 1)

    def test_assignment_manipulation(self):
        self.ar.assignment.amount = 2
        self.ar.save()
        self.assertEqual(self.ar.assignment.amount, 1)



# Generated by Django 2.2.4 on 2019-10-22 21:53

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('juntagrico_assignment_request', '0001_squashed_0006_auto_20191020_2203'),
    ]

    operations = [
        migrations.AlterField(
            model_name='assignmentrequest',
            name='assignment',
            field=models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='juntagrico.Assignment'),
        ),
        migrations.AlterField(
            model_name='assignmentrequest',
            name='activityarea',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='juntagrico.ActivityArea', verbose_name='Tätigkeitsbereich'),
        ),
        migrations.AlterField(
            model_name='assignmentrequest',
            name='approver',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='Referenz', to='juntagrico.Member', verbose_name='Abgesprochen mit'),
        ),
        migrations.AlterField(
            model_name='assignmentrequest',
            name='assignment',
            field=models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='juntagrico.Assignment', verbose_name='Arbeitseinsatz'),
        ),
        migrations.AlterField(
            model_name='assignmentrequest',
            name='member',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='juntagrico.Member', verbose_name='Mitglied'),
        ),
    ]



from django.forms import ModelForm, URLField
from django.urls import reverse

from juntagrico.admins import BaseAdmin
from juntagrico.util.admin import MyHTMLWidget
from juntagrico.config import Config

from juntagrico_assignment_request.dao.assignmentrequestdao import AssignmentRequestDao
from juntagrico_assignment_request.entity.assignment_request import AssignmentRequest


class AssignmentRequestAdminForm(ModelForm):
    class Meta:
        model = AssignmentRequest
        fields = '__all__'

    def __init__(self, *a, **k):
        super().__init__(*a, **k)
        self.fields['approver'].queryset = AssignmentRequestDao.all_approvers()
        instance = k.get('instance')
        self.fields['assignment_link'].initial = self.get_assignment_link(instance)

    @staticmethod
    def get_assignment_link(instance):
        if instance is None:
            return ''
        elif instance.assignment:
            return '<a href={}>{}</a>'.format(
                reverse('admin:juntagrico_assignment_change', args=(instance.assignment.id,)),
                instance.assignment
            )
        else:
            return '-'

    assignment_link = URLField(widget=MyHTMLWidget(), required=False, label=Config.vocabulary('assignment'))


class AssignmentRequestAdmin(BaseAdmin):
    form = AssignmentRequestAdminForm
    raw_id_fields = ['member']
    list_display = ['member', 'amount', 'job_time', 'duration', 'activityarea',
                    'location', 'request_date', 'approver', 'status', ]
    exclude = ('assignment',)



from django.db.models import signals

from juntagrico_assignment_request.entity.assignment_request import *

signals.pre_save.connect(AssignmentRequest.pre_save, sender=AssignmentRequest)


def bulk_save(elements):
    for element in elements:
        element.save()
        print(f'Resaved {element}')


def post_save_job_type(sender, instance, **kwds):
    """
    As we can not prevent admins from messing with autogenerated job types, we
    re-save all assignment requests. if any relevant part of this job type was changed,
    the requests will create a new matching job type and adapt their jobs to that.
    """
    # get assignment requests that are attached to any assignments of jobs of that job type and re-save them
    bulk_save(AssignmentRequest.objects.filter(
        assignment__job__in=instance.recuringjob_set.filter(assignment__assignmentrequest__isnull=False)
    ))


def post_save_recuringjob(sender, instance, **kwds):
    """
    same idea as post_save_job_type
    """
    # get assignment requests that are attached to any assignments of this job and re-save them
    bulk_save(AssignmentRequest.objects.filter(
        assignment__in=instance.assignment_set.filter(assignmentrequest__isnull=False)
    ))


def pre_save_assignment(sender, instance, **kwds):
    """
    If assignment comes from request, change back values as before saving it
    """
    if hasattr(instance, 'assignmentrequest'):
        ar = instance.assignmentrequest
        instance.member = ar.member
        instance.amount = ar.amount
        instance.job = ar.get_matching_job()
        print(f'restored {instance} using {ar}')


signals.post_save.connect(post_save_job_type, sender=JobType)
signals.post_save.connect(post_save_recuringjob, sender=RecuringJob)
signals.pre_save.connect(pre_save_assignment, sender=Assignment)



"""juntagrico_crowdfunding URL Configuration
"""
from django.urls import path
from juntagrico_assignment_request import views as assignment_request

urlpatterns = [
    path('ar/assignment/request', assignment_request.request_assignment, name='ar-request-assignment'),
    path('ar/assignment/requested', assignment_request.request_assignment,
         {'sent': True}, name='ar-assignment-requested'),
    path('ar/assignment/delete/<int:request_id>/', assignment_request.delete_request_assignment,
         name='ar-delete-assignment-request'),
    path('ar/assignment/edit/<int:request_id>/', assignment_request.edit_request_assignment,
         name='ar-edit-assignment-request'),
    path('ar/assignment/list', assignment_request.list_assignment_requests,
         name='ar-list-assignment-requests'),
    path('ar/assignment/archive', assignment_request.list_archive, name='ar-list-archive'),
    path('ar/assignment/respond/<int:request_id>/', assignment_request.respond_assignment_request,
         name='ar-respond-assignment-request'),
    path('ar/assignment/confirm/<int:request_id>/', assignment_request.confirm_assignment_request,
         name='ar-confirm-assignment-request'),
]



from django.core import mail
from django.urls import reverse

from juntagrico_assignment_request.entity.assignment_request import AssignmentRequest
from test import AssignmentRequestTestCase


class AssignmentRequestTests(AssignmentRequestTestCase):
    def test_assignment_request(self):
        self.assertGet(reverse('ar-request-assignment'))
        self.assertPost(reverse('ar-request-assignment'),
                        self.assignment_request_data(self.approver, for_form=True), 302)
        ar = AssignmentRequest.objects.filter(approver=self.approver.pk)
        self.assertEqual(ar.count(), 1)
        self.assertEqual(ar.first().status, AssignmentRequest.REQUESTED)
        self.assertEqual(len(mail.outbox), 1)  # request email to approver
        self.assertEqual(mail.outbox[0].to[0], self.approver.email)

    def test_assignment_request_edit(self):
        ar = AssignmentRequest.objects.create(**self.assignment_request_data(self.approver))
        self.assertGet(reverse('ar-request-assignment'))  # test list of existing own requests
        self.assertGet(reverse('ar-edit-assignment-request', args=(ar.pk,)))
        self.assertPost(reverse('ar-edit-assignment-request', args=(ar.pk,)),
                        self.assignment_request_data(self.approver, for_form=True), 302)
        ar.refresh_from_db()
        self.assertEqual(ar.status, AssignmentRequest.REQUESTED)
        self.assertEqual(len(mail.outbox), 1)  # edit notification to approver
        self.assertEqual(mail.outbox[0].to[0], self.approver.email)

    def test_assignment_request_delete(self):
        ar = AssignmentRequest.objects.create(**self.assignment_request_data(self.approver))
        self.assertGet(reverse('ar-delete-assignment-request', args=(ar.pk,)), 302)
        with self.assertRaises(AssignmentRequest.DoesNotExist):
            ar.refresh_from_db()

    def test_assignment_request_not_deletable(self):
        # can not delete accepted assignment request
        data = self.assignment_request_data(self.approver, approved=True)
        ar = AssignmentRequest.objects.create(**data)
        self.assertGet(reverse('ar-delete-assignment-request', args=(ar.pk,)), 302)
        ar.refresh_from_db()
        self.assertNotEquals(ar, None)

    def test_assignment_request_wo_approver(self):
        self.assertPost(reverse('ar-request-assignment'), self.assignment_request_data(for_form=True), 302)
        ar = AssignmentRequest.objects.filter(approver=self.approver.pk)
        self.assertEqual(ar.count(), 0)
        self.assertEqual(len(mail.outbox), 1)  # request email to general approver
        self.assertEqual(mail.outbox[0].to[0], self.approver2.email)

    def test_assignment_request_list(self):
        AssignmentRequest.objects.create(**self.assignment_request_data(self.approver))  # display at least one ar
        self.assertGet(reverse('ar-list-assignment-requests'), 302)  # normal member has no access
        self.assertGet(reverse('ar-list-assignment-requests'), member=self.approver)

    def test_assignment_request_archive(self):
        AssignmentRequest.objects.create(**self.assignment_request_data(self.approver))  # display at least one ar
        self.assertGet(reverse('ar-list-archive'), 302)  # normal member has no access
        self.assertGet(reverse('ar-list-archive'), member=self.approver)

    def test_assignment_confirmation(self):
        ar = AssignmentRequest.objects.create(**self.assignment_request_data(self.approver))
        self.assertGet(reverse('ar-confirm-assignment-request', args=(ar.pk,)), 302)  # member can not approve
        ar.refresh_from_db()
        self.assertEqual(ar.status, AssignmentRequest.REQUESTED)
        self.assertGet(reverse('ar-confirm-assignment-request', args=(ar.pk,)), 302, member=self.approver)
        ar.refresh_from_db()
        self.assertEqual(ar.status, AssignmentRequest.CONFIRMED)
        self.assertEqual(len(mail.outbox), 1)  # confirmation email to user
        self.assertEqual(mail.outbox[0].to[0], self.member.email)

    def test_assignment_confirmation_with_response_by_other_approver(self):
        ar = AssignmentRequest.objects.create(**self.assignment_request_data(self.approver))
        self.assertGet(reverse('ar-respond-assignment-request', args=(ar.pk,)), member=self.approver)
        self.assertPost(reverse('ar-respond-assignment-request', args=(ar.pk,)),
                        self.assignment_response_data('confirm'), 302, member=self.approver2)
        ar.refresh_from_db()
        self.assertEqual(ar.status, AssignmentRequest.CONFIRMED)
        self.assertEqual(ar.approver, self.approver2)  # approver changed to actual approver
        self.assertEqual(len(mail.outbox), 2)  # rejection email to user and information to original approver
        self.assertEqual(mail.outbox[0].to[0], self.approver.email)
        self.assertEqual(mail.outbox[1].to[0], self.member.email)

    def assignment_response(self, decision='submit', expected_status=AssignmentRequest.REQUESTED):
        ar = AssignmentRequest.objects.create(**self.assignment_request_data(self.approver))
        self.assertPost(reverse('ar-respond-assignment-request', args=(ar.pk,)),
                        self.assignment_response_data(decision), 302, member=self.approver)
        ar.refresh_from_db()
        self.assertEqual(ar.status, expected_status)
        self.assertEqual(len(mail.outbox), 1)  # reply email to user
        self.assertEqual(mail.outbox[0].to[0], self.member.email)

    def test_assignment_reply(self):
        self.assignment_response()

    def test_assignment_rejection(self):
        self.assignment_response('reject', AssignmentRequest.REJECTED)
